Hex.TYPE_PLAIN = 0;
Hex.TYPE_STREET = 1;
Hex.TYPE_PALACE = 2;
Hex.TYPE_TOWN = 3;
Hex.TYPE_COAST = 4;
Hex.TYPE_SEA = 5;
Hex.TYPE_DINGHI = 6;
Hex.MODE_INFRA = 0;
Hex.MODE_ARM = 1;
Hex.MODE_DEFEND = 2;

function Hex(u, v, type) {
    this.u = u;
    this.v = v;
    this.type = type;
    this.friendCount = 0;
    this.enemyCount = 0;
}

Hex.prototype["createElement"] = function hexCreateElement(parent, board) {
    this.board = board;
    var element = document.createElement("div");
    var style = "position: absolute; left: " + (u * 45) + "px; "
      + "top: " + ((v - Math.floor(u / 2)) * 52 + (isEven(u)? 1 : 0) * 26) + "px;";
    element.setAttribute("style", style);
    //element.setAttribute("onMouseOver", "setMessage('" + u + ", " + v + "');");
    element.setAttribute("onClick", "document.board.advanceTurn(" + u + ", " + v + ");");
    var img = document.createElement("img");
    img.setAttribute("src", this.getBgTileName());
    element.appendChild(img);
    if (this.type == Hex.TYPE_TOWN) {
        var sprite = document.createElement("img");
        sprite.setAttribute("src", "gifs/stadt.gif");
        sprite.setAttribute("style", "position: absolute; left: 0px; top: 0px;");
        element.appendChild(sprite);
        this.sizeElement = document.createElement("p");
        var description = (this.intact)? ("" + this.size) : "0";
        this.sizeElement.appendChild(document.createTextNode(description));
        this.sizeElement.setAttribute("style", "position: absolute; top: 0px; left: 20px;");
        element.appendChild(this.sizeElement);
    }
    this.unitElement = document.createElement("div");
    element.appendChild(this.unitElement);
    this.unitElement.setAttribute("style", "position: absolute; top: 0px; left: 0px;");
    this.redrawUnits();
    parent.appendChild(element);
    return element;
}

Hex.prototype["getTypeString"] = function hexGetTypeString() {
    switch (this.type) {
        case Hex.TYPE_PLAIN:
          return "Plain";
        case Hex.TYPE_STREET:
          return "Street";
        case Hex.TYPE_PALACE:
          return "Palace";
        case Hex.TYPE_TOWN:
          return "Town";
        case Hex.TYPE_COAST:
          return "Coast";
        case Hex.TYPE_SEA:
          return "Sea";
        case Hex.TYPE_DINGHI:
          return "Dinghi";
    }
}

Hex.prototype["getBgTileName"] = function hexGetBgTileName() {
    switch (this.type) {
        case Hex.TYPE_PLAIN:
          return "gifs/wueste.gif";
        case Hex.TYPE_STREET:
          return "gifs/strasse.gif";
        case Hex.TYPE_PALACE:
          return "gifs/palast.gif";
        case Hex.TYPE_TOWN:
          return "gifs/strasse.gif";
        case Hex.TYPE_COAST:
          return "gifs/kueste.gif";
        case Hex.TYPE_SEA:
          return "gifs/meer.gif";
        case Hex.TYPE_DINGHI:
          return "gifs/meerBoot.gif";
    }
}

Hex.prototype["getDistanceTo"] = function hexGetDistanceTo(otherHex) {
    var dU = otherHex.u - this.u;
    var dV = otherHex.v - this.v;
    if (dU >= 0) {
        if (dV >= 0) {
            return Math.max(dU, dV);
        } else {
            return dU + dV * (-1);
        }
    } else {
        if (dV <= 0) {
            return Math.min(dU, dV) * (-1);
        } else {
            return dU * (-1) + dV;
        }
    }
}

Hex.prototype["advanceTurn"] = function hexAdvanceTurn() {
    if (this.type == Hex.TYPE_TOWN) {
        if ((this.mode == Hex.MODE_INFRA) && this.intact) {
            this.size += 1;
            this.sizeElement.removeChild(this.sizeElement.childNodes[0]);
            this.sizeElement.appendChild(document.createTextNode("" + this.size));
        } else if ((this.mode == Hex.MODE_ARM) && this.intact) {
            this.friendCount += this.size;
            this.redrawUnits();
        }
    } else if (this.type == Hex.TYPE_DINGHI) {
        
    }
}

Hex.prototype["toggle"] = function hexToggle() {
    if (this.type == Hex.TYPE_TOWN) {
        if (this.mode == Hex.MODE_INFRA) {
            this.mode = Hex.MODE_ARM;
            return true;
        } else if (this.mode == Hex.MODE_ARM) {
            this.mode = Hex.MODE_DEFEND;
            return true;
        }
    }
    return false;
}

Hex.prototype["redrawUnits"] = function hexRedrawUnits() {
    while (this.unitElement.hasChildNodes()) {
        this.unitElement.removeChild(this.unitElement.firstChild);
    }
    if ((this.friendCount == 0) && (this.enemyCount == 0)) {
        return;
    }
    if (this.friendCount > 0) {
        var spriteImgName = "gifs/soldatE.gif";
        var count = this.friendCount;
    } else {
        var spriteImgName = "gifs/soldatG.gif";
        var count = this.enemyCount;        
    }
    var sprite = document.createElement("img");
    sprite.setAttribute("src", spriteImgName);
    this.unitElement.appendChild(sprite);
    var countTxt = document.createElement("p");
    countTxt.appendChild(document.createTextNode("" + count));
    countTxt.setAttribute("style", "position: absolute; top: 5px; left: 28px;");
    this.unitElement.appendChild(countTxt);
}

function Board(width, height, randomSeed) {
    this.width = width;
    this.height = height;
    this.turn = 1;
    this.random = new Random(randomSeed);
    this.noTowns = Math.round(width * height / 12);
    this.noDinghis = 2;
    this.enemyUnit = new Array();
    this.friendlyUnit = new Array();
    this.maxV = Math.floor(width / 2) + height;
    this.hex = new Array(width);
    for (u = 0; u < width; u++) {
        this.hex[u] = new Array(this.maxV);
        var halfU = Math.floor(u / 2);
        var h = isEven(u)? 0 : 1;
        for (v = halfU; v < halfU + height + h - 1; v++) {
            this.hex[u][v] = new Hex(u, v, Hex.TYPE_PLAIN);
        }
    }
    setMessage("level number: " + randomSeed, "lvlNo");
    setMessage("turn: " + this.turn, "turnNo");
    document.board = this;
}

Board.prototype["createElement"] = function boardCreateElement(parent) {
    var par = document.createElement("p");
    parent.appendChild(par);
    for (u = 0; u < this.width; u++) {
        for (v = 0; v <= this.maxV; v++) {
            var h = this.hex[u][v];
            h && h.createElement(par);
        }
    }
    return par;
}

Board.prototype["initMap"] = function boardInitMap() {
    this.setColumnToType(this.width - 1, Hex.TYPE_SEA);
    this.setColumnToType(this.width - 2, Hex.TYPE_COAST);
    palaceV = Math.round(this.random.next() * (this.height - 2));
    this.hex[0][palaceV].type = Hex.TYPE_PALACE;
    var minTownU = 1;
    var maxTownU = this.width - 3;
    var i = 0;
    var metaTries = 0;
    while ((i < this.noTowns) && (metaTries < 128)) {
        this.clearTowns();
        for (i = 0, tries = 0; (i < this.noTowns) && (tries < 255); tries++) {
            var townU = minTownU + Math.round((maxTownU - minTownU) * this.random.next());
            var minTownV = Math.floor(townU / 2);
            var h = isEven(townU)? 0 : 1;
            var maxTownV = minTownV + this.height + h - 2;
            var townV = minTownV + Math.round((maxTownV - minTownV) * this.random.next());
            var potentialTownHex = this.hex[townU][townV];
            if ((potentialTownHex.type != Hex.TYPE_TOWN)
                  && (this.isTownPlacementOk(potentialTownHex))) {
                potentialTownHex.type = Hex.TYPE_TOWN;
                potentialTownHex.size = 1;
                potentialTownHex.intact = true;
                potentialTownHex.mode = Hex.MODE_INFRA;
                i++;
            }
        }
        metaTries++;
    }
    for (i = 0, tries = 0; (i < this.noDinghis) && (tries < 255); tries++) {
        var dinghiU = this.width - 1;
        var minDinghiV = Math.floor(dinghiU / 2);
        var h = isEven(dinghiU)? 0 : 1;
        var maxDinghiV = minDinghiV + this.height + h - 2;
        var dinghiV = minDinghiV + Math.round((maxDinghiV - minDinghiV) * this.random.next());
        var potentialDinghiHex = this.hex[dinghiU][dinghiV];
        if ((potentialDinghiHex.type != Hex.TYPE_DINGHI)
              && (this.isDinghiPlacementOk(potentialDinghiHex))) {
            potentialDinghiHex.type = Hex.TYPE_DINGHI;
            potentialDinghiHex.enemyCount = 30;
            i++;
        }
    }
}

Board.prototype["setColumnToType"] = function boardSetColumnToType(column, type) {
    for (v = 0; v <= this.maxV; v++) {
        var tile = this.hex[column][v];
        tile && (tile.type = type);
    }
}

Board.prototype["getAllTowns"] = function boardGetAllTowns() {
    var retVal = new Array();
    var i = 0;
    for (u = 0; u < this.width; u++) {
        for (v = 0; v <= this.maxV; v++) {
            var h = this.hex[u][v];
            if (h && h.type == Hex.TYPE_TOWN) {
                retVal[i] = h;
                i++;
            }
        }
    }
    return retVal;
}

Board.prototype["getAllDinghis"] = function boardGetAllDinghis() {
    var retVal = new Array();
    var i = 0;
    for (u = 0; u < this.width; u++) {
        for (v = 0; v <= this.maxV; v++) {
            var h = this.hex[u][v];
            if (h && h.type == Hex.TYPE_DINGHI) {
                retVal[i] = h;
                i++;
            }
        }
    }
    return retVal;
}

Board.prototype["clearTowns"] = function boardClearTowns() {
    for (u = 0; u < this.width; u++) {
        for (v = 0; v <= this.maxV; v++) {
            var h = this.hex[u][v];
            if (h && h.type == Hex.TYPE_TOWN) {
                h.type = Hex.TYPE_PLAIN;
            }
        }
    }    
}

Board.prototype["isTownPlacementOk"] = function boardIsTownPlacementOk(hex) {
    var allTowns = this.getAllTowns();
    for (i = 0; i < allTowns.length; i++) {
        if (hex.getDistanceTo(allTowns[i]) <= 1) {
            return false;
        }
    }
    return true;
}

Board.prototype["isDinghiPlacementOk"] = function boardIsDinghiPlacementOk(hex) {
    var allDinghis = this.getAllDinghis();
    for (i = 0; i < allDinghis.length; i++) {
        if (hex.getDistanceTo(allDinghis[i]) <= 2) {
            return false;
        }
    }
    return true;
}

Board.prototype["advanceTurn"] = function boardAdvanceTurn(u, v) {
    var commence = true;
    if ((u != undefined) && (v != undefined)) {
        commence = this.hex[u][v].toggle();
    }
    if (!commence) {
        return;
    }
    this.turn += 1;
    setMessage("turn: " + this.turn, "turnNo");
    for (u = 0; u < this.width; u++) {
        for (v = 0; v <= this.maxV; v++) {
            var h = this.hex[u][v];
            if (h) {
                h.advanceTurn();
            }
        }
    }
}

function initialise() {
    /*var page = window.document.body;
    var paragraph = document.createElement("p");
    paragraph.appendChild(document.createTextNode("paragraph"));
    page.appendChild(paragraph);*/
    var board = new Board(13, 8, Math.floor(Math.random() * 65000));
    board.initMap();
    board.createElement(window.document.body);
}

function setMessage(message, type) {
    var msgElement = document.getElementById(type);
    /*var blah = "";
    for (p in msgElement.childNodes) {
        blah = blah + " " + p;
    }*/
    msgElement.removeChild(msgElement.childNodes[0]);
    msgElement.appendChild(document.createTextNode(message));
}

function isEven(number) {
    return Math.floor(number / 2) == (number / 2);
}

Random.prototype.RANGE = Math.pow(2, 16);

function Random(seed) {
    this.metaRandom = new MetaRandom(seed);
}

Random.prototype["next"] = function randomNext() {
    var bit = new Array(16);
    for (var i = 0; i < 16; i++) {
        bit[i] = (isEven(this.metaRandom.next()))? 0 : 1;
    }
    var h = 1;
    var rndNum = 0;
    for (i = 0; i < 16; i++) {
        if (bit[i]) {
            rndNum = rndNum + h;
        }
        h = h * 2;
    }
    return rndNum / this.RANGE;
}

function MetaRandom(seed) {
    this.last = seed;
}

MetaRandom.prototype["next"] = function metaRandomNext() {
    var message = "";
    var bit = new Array(16);
    // get last number's bit pattern
    for (var i = 0; i < 16; i++) {
        bit[i] = (isEven(this.last))? 0 : 1;
        this.last = Math.floor(this.last / 2);
        message = message + bit[i];
    }
    // save first bit for later
    var b0 = bit[0];
    // shift the bits around
    for (i = 0; i < 15; i++) {
        bit[i] = bit[i + 1];
    }
    // fill in the first bit at the other end
    bit[15] = b0;
    // save some of the bits temporarily for the next step
    b0 = bit[1];
    var b1 = bit[3];
    var b2 = bit[4];
    // now twiddle them around a bit
    bit[1] = (b0 != b1);
    bit[3] = (b1 != b2);
    bit[4] = (b0 != b2);
    // convert the bit pattern into a number again
    var h = 1;
    var next = 0;
    for (i = 0; i < 16; i++) {
        if (bit[i]) {
            next = next + h;
        }
        h = h * 2;
    }
    //setMessage(message + " " + next);
    this.last = next;
    return next;
}